@@ -1,16 +1,18 @@
 import json
 import logging
 import os
+import re
 import subprocess
 from contextlib import suppress
 from copy import copy
 from os import mkdir
-from os.path import basename, dirname, exists, join, splitext
+from os.path import basename, dirname, exists, join, relpath, splitext
 from shutil import rmtree
-from typing import Any, Dict
+from typing import Any, Dict, cast
 
 from jinja2 import Template
 
+from dipdup import __version__
 from dipdup.config import (
     CONFIGURE_HANDLER,
     ROLLBACK_HANDLER,
@@ -27,7 +29,22 @@ from dipdup.config import (
 from dipdup.datasources import DatasourceT
 from dipdup.datasources.tzkt.datasource import TzktDatasource
 from dipdup.exceptions import ConfigurationError
-from dipdup.utils import pascal_to_snake, snake_to_pascal
+from dipdup.utils import import_submodules, pascal_to_snake, snake_to_pascal
+
+DEFAULT_DOCKER_ENV_FILE_CONTENT = dict(
+    POSTGRES_USER="dipdup",
+    POSTGRES_DB="dipdup",
+    POSTGRES_PASSWORD="changeme",
+    HASURA_GRAPHQL_DATABASE_URL="postgres://dipdup:changeme@db:5432/dipdup",
+    HASURA_GRAPHQL_ENABLE_CONSOLE="true",
+    HASURA_GRAPHQL_DEV_MODE="true",
+    HASURA_GRAPHQL_ENABLED_LOG_TYPES="startup, http-log, webhook-log, websocket-log, query-log",
+    HASURA_GRAPHQL_ADMIN_SECRET="changeme",
+    HASURA_GRAPHQL_UNAUTHORIZED_ROLE="user",
+)
+DEFAULT_DOCKER_IMAGE = 'dipdup/dipdup'
+DEFAULT_DOCKER_TAG = __version__
+DEFAULT_DOCKER_ENV_FILE = 'dipdup.env'
 
 
 def resolve_big_maps(schema: Dict[str, Any]) -> Dict[str, Any]:
@@ -48,11 +65,25 @@ class DipDupCodeGenerator:
     """Generates package based on config, invoked from `init` CLI command"""
 
     def __init__(self, config: DipDupConfig, datasources: Dict[DatasourceConfigT, DatasourceT]) -> None:
-        self._logger = logging.getLogger(__name__)
+        self._logger = logging.getLogger('dipdup.codegen')
         self._config = config
         self._datasources = datasources
         self._schemas: Dict[TzktDatasourceConfig, Dict[str, Dict[str, Any]]] = {}
 
+    async def init(self) -> None:
+        await self.create_package()
+        await self.fetch_schemas()
+        await self.generate_types()
+        await self.generate_default_handlers()
+        await self.generate_user_handlers()
+        await self.generate_jobs()
+        await self.cleanup()
+        await self.verify_package()
+
+    async def docker_init(self, image: str, tag: str, env_file: str) -> None:
+        await self.generate_docker(image, tag, env_file)
+        await self.verify_package()
+
     async def create_package(self) -> None:
         """Create Python package skeleton if not exists"""
         self._logger.info('Creating package `%s`', self._config.package)
@@ -80,6 +111,13 @@ class DipDupCodeGenerator:
             with open(join(handlers_path, '__init__.py'), 'w'):
                 pass
 
+        self._logger.info('Creating `%s.jobs` package', self._config.package)
+        jobs_path = join(self._config.package_path, 'jobs')
+        with suppress(FileExistsError):
+            mkdir(jobs_path)
+            with open(join(jobs_path, '__init__.py'), 'w'):
+                pass
+
         self._logger.info('Creating `%s/sql` directory', self._config.package)
         sql_path = join(self._config.package_path, 'sql')
         with suppress(FileExistsError):
@@ -87,9 +125,20 @@ class DipDupCodeGenerator:
         sql_on_restart_path = join(sql_path, 'on_restart')
         with suppress(FileExistsError):
             mkdir(sql_on_restart_path)
+        with open(join(sql_on_restart_path, '.keep'), 'w'):
+            pass
         sql_on_reindex_path = join(sql_path, 'on_reindex')
         with suppress(FileExistsError):
             mkdir(sql_on_reindex_path)
+        with open(join(sql_on_reindex_path, '.keep'), 'w'):
+            pass
+
+        self._logger.info('Creating `%s/graphql` directory', self._config.package)
+        graphql_path = join(self._config.package_path, 'graphql')
+        with suppress(FileExistsError):
+            mkdir(graphql_path)
+        with open(join(graphql_path, '.keep'), 'w'):
+            pass
 
     async def fetch_schemas(self) -> None:
         """Fetch JSONSchemas for all contracts used in config"""
@@ -134,21 +183,19 @@ class DipDupCodeGenerator:
                             continue
 
                         parameter_schemas_path = join(contract_schemas_path, 'parameter')
+                        entrypoint = cast(str, operation_pattern_config.entrypoint)
                         with suppress(FileExistsError):
                             mkdir(parameter_schemas_path)
 
                         try:
                             entrypoint_schema = next(
-                                ep['parameterSchema']
-                                for ep in contract_schemas['entrypoints']
-                                if ep['name'] == operation_pattern_config.entrypoint
+                                ep['parameterSchema'] for ep in contract_schemas['entrypoints'] if ep['name'] == entrypoint
                             )
                         except StopIteration as e:
-                            raise ConfigurationError(
-                                f'Contract `{contract_config.address}` has no entrypoint `{operation_pattern_config.entrypoint}`'
-                            ) from e
+                            raise ConfigurationError(f'Contract `{contract_config.address}` has no entrypoint `{entrypoint}`') from e
 
-                        entrypoint_schema_path = join(parameter_schemas_path, f'{operation_pattern_config.entrypoint}.json')
+                        entrypoint = entrypoint.replace('.', '_').lstrip('_')
+                        entrypoint_schema_path = join(parameter_schemas_path, f'{entrypoint}.json')
 
                         if not exists(entrypoint_schema_path):
                             with open(entrypoint_schema_path, 'w') as file:
@@ -181,15 +228,16 @@ class DipDupCodeGenerator:
                         raise ConfigurationError(
                             f'Contract `{contract_config.address}` has no big map path `{big_map_handler_config.path}`'
                         ) from e
+                    big_map_path = big_map_handler_config.path.replace('.', '_')
                     big_map_key_schema = big_map_schema['keySchema']
-                    big_map_key_schema_path = join(big_map_schemas_path, f'{big_map_handler_config.path}.key.json')
+                    big_map_key_schema_path = join(big_map_schemas_path, f'{big_map_path}_key.json')
 
                     if not exists(big_map_key_schema_path):
                         with open(big_map_key_schema_path, 'w') as file:
                             file.write(json.dumps(big_map_key_schema, indent=4))
 
                     big_map_value_schema = big_map_schema['valueSchema']
-                    big_map_value_schema_path = join(big_map_schemas_path, f'{big_map_handler_config.path}.value.json')
+                    big_map_value_schema_path = join(big_map_schemas_path, f'{big_map_path}_value.json')
 
                     if not exists(big_map_value_schema_path):
                         with open(big_map_value_schema_path, 'w') as file:
@@ -230,6 +278,12 @@ class DipDupCodeGenerator:
                 input_path = join(root, file)
                 output_path = join(types_root, f'{pascal_to_snake(name)}.py')
 
+                if exists(output_path):
+                    with open(output_path) as type_file:
+                        first_line = type_file.readline()
+                        if re.match(r'^#\s+dipdup:\s+ignore\s*', first_line):
+                            continue
+
                 if name == 'storage':
                     name = '_'.join([root.split('/')[-1], name])
                 if root.split('/')[-1] == 'parameter':
@@ -244,7 +298,7 @@ class DipDupCodeGenerator:
                     '--output',
                     output_path,
                     '--class-name',
-                    name,
+                    name.lstrip('_'),
                     '--disable-timestamp',
                     '--use-default',
                 ]
@@ -253,9 +307,9 @@ class DipDupCodeGenerator:
 
     async def generate_default_handlers(self, recreate=False) -> None:
         handlers_path = join(self._config.package_path, 'handlers')
-        with open(join(dirname(__file__), 'templates', f'{ROLLBACK_HANDLER}.py.j2')) as file:
+        with open(join(dirname(__file__), 'templates', 'handlers', f'{ROLLBACK_HANDLER}.py.j2')) as file:
             rollback_template = Template(file.read())
-        with open(join(dirname(__file__), 'templates', f'{CONFIGURE_HANDLER}.py.j2')) as file:
+        with open(join(dirname(__file__), 'templates', 'handlers', f'{CONFIGURE_HANDLER}.py.j2')) as file:
             configure_template = Template(file.read())
 
         self._logger.info('Generating handler `%s`', CONFIGURE_HANDLER)
@@ -275,9 +329,9 @@ class DipDupCodeGenerator:
     async def generate_user_handlers(self) -> None:
         """Generate handler stubs with typehints from templates if not exist"""
         handlers_path = join(self._config.package_path, 'handlers')
-        with open(join(dirname(__file__), 'templates', 'operation_handler.py.j2')) as file:
+        with open(join(dirname(__file__), 'templates', 'handlers', 'operation.py.j2')) as file:
             operation_handler_template = Template(file.read())
-        with open(join(dirname(__file__), 'templates', 'big_map_handler.py.j2')) as file:
+        with open(join(dirname(__file__), 'templates', 'handlers', 'big_map.py.j2')) as file:
             big_map_handler_template = Template(file.read())
 
         for index_config in self._config.indexes.values():
@@ -315,12 +369,94 @@ class DipDupCodeGenerator:
             else:
                 raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
 
+    async def generate_jobs(self) -> None:
+        if not self._config.jobs:
+            return
+
+        jobs_path = join(self._config.package_path, 'jobs')
+        with open(join(dirname(__file__), 'templates', 'jobs', 'job.py.j2')) as file:
+            job_template = Template(file.read())
+
+        job_callbacks = set(job_config.callback for job_config in self._config.jobs.values())
+        for job_callback in job_callbacks:
+            self._logger.info('Generating job `%s`', job_callback)
+            job_code = job_template.render(job=job_callback)
+            job_path = join(jobs_path, f'{job_callback}.py')
+            if not exists(job_path):
+                with open(job_path, 'w') as file:
+                    file.write(job_code)
+
+    async def generate_docker(self, image: str, tag: str, env_file: str) -> None:
+        self._logger.info('Generating Docker template')
+        docker_path = join(self._config.package_path, 'docker')
+        with suppress(FileExistsError):
+            mkdir(docker_path)
+
+        with open(join(dirname(__file__), 'templates', 'docker', 'Dockerfile.j2')) as file:
+            dockerfile_template = Template(file.read())
+        with open(join(dirname(__file__), 'templates', 'docker', 'docker-compose.yml.j2')) as file:
+            docker_compose_template = Template(file.read())
+        with open(join(dirname(__file__), 'templates', 'docker', 'dipdup.env.j2')) as file:
+            dipdup_env_template = Template(file.read())
+
+        dockerfile_code = dockerfile_template.render(
+            image=f'{image}:{tag}',
+            package=self._config.package,
+            package_path=self._config.package_path,
+        )
+        dockerfile_path = join(docker_path, 'Dockerfile')
+        with open(dockerfile_path, 'w') as file:
+            file.write(dockerfile_code)
+
+        mounts = {}
+        for filename in self._config.filenames:
+            filename_part = filename.split("/")[-1]
+            from_ = join(relpath(self._config.package_path, filename), filename_part)
+            to = f'/home/dipdup/{filename_part}'
+            mounts[from_] = to
+
+        command = []
+        for filename in self._config.filenames:
+            command += ['-c', filename.split("/")[-1]]
+        command += ['run']
+
+        docker_compose_code = docker_compose_template.render(
+            package=self._config.package,
+            mounts=mounts,
+            env_file=env_file,
+            command=command,
+        )
+        docker_compose_path = join(docker_path, 'docker-compose.yml')
+        with open(docker_compose_path, 'w') as file:
+            file.write(docker_compose_code)
+
+        dipdup_env_code = dipdup_env_template.render(
+            environment={
+                **DEFAULT_DOCKER_ENV_FILE_CONTENT,
+                **self._config.environment,
+            }
+        )
+        dipdup_env_example_path = join(docker_path, f'{env_file}.example')
+        with open(dipdup_env_example_path, 'w') as file:
+            file.write(dipdup_env_code)
+        dipdup_env_path = join(docker_path, env_file)
+        if not exists(dipdup_env_path):
+            with open(dipdup_env_path, 'w') as file:
+                file.write(dipdup_env_code)
+
+        gitignore_path = join(docker_path, '.gitignore')
+        with open(gitignore_path, 'w') as file:
+            file.write('*.env')
+
     async def cleanup(self) -> None:
         """Remove fetched JSONSchemas"""
         self._logger.info('Cleaning up')
         schemas_path = join(self._config.package_path, 'schemas')
         rmtree(schemas_path)
 
+    async def verify_package(self) -> None:
+        import_submodules(self._config.package)
+
     async def _get_schema(
         self,
         datasource_config: TzktDatasourceConfig,
@@ -347,7 +483,7 @@ class DipDupCodeGenerator:
             self._schemas[datasource_config][address] = address_schemas_json
         return self._schemas[datasource_config][address]
 
-    async def migrate_user_handlers_to_v1(self) -> None:
+    async def migrate_user_handlers_to_v10(self) -> None:
         remove_lines = [
             'from dipdup.models import',
             'from dipdup.context import',
@@ -383,3 +519,26 @@ class DipDupCodeGenerator:
                         newfile.append(line)
                 with open(path, 'w') as file:
                     file.write('\n'.join(newfile))
+
+    async def migrate_user_handlers_to_v11(self) -> None:
+        replace_table = {
+            'BigMapAction.ADD': 'BigMapAction.ADD_KEY',
+            'BigMapAction.UPDATE': 'BigMapAction.UPDATE_KEY',
+            'BigMapAction.REMOVE': 'BigMapAction.REMOVE_KEY',
+        }
+        handlers_path = join(self._config.package_path, 'handlers')
+
+        for root, _, files in os.walk(handlers_path):
+            for filename in files:
+                if filename == '__init__.py' or not filename.endswith('.py'):
+                    continue
+                path = join(root, filename)
+                newfile = []
+                with open(path) as file:
+                    for line in file.read().split('\n'):
+                        # Replace by table
+                        for from_, to in replace_table.items():
+                            line = line.replace(from_, to)
+                        newfile.append(line)
+                with open(path, 'w') as file:
+                    file.write('\n'.join(newfile))
