@@ -26,9 +26,11 @@ ROLLBACK_HANDLER = 'on_rollback'
 CONFIGURE_HANDLER = 'on_configure'
 BLOCK_HANDLER = 'on_block'
 ENV_VARIABLE_REGEX = r'\${([\w]*):-(.*)}'
+DEFAULT_RETRY_COUNT = 3
+DEFAULT_RETRY_SLEEP = 1
 
 sys.path.append(os.getcwd())
-_logger = logging.getLogger(__name__)
+_logger = logging.getLogger('dipdup.config')
 
 
 class OperationType(Enum):
@@ -76,7 +78,9 @@ class PostgresDatabaseConfig:
 
     @property
     def connection_string(self) -> str:
-        return f'{self.kind}://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}?schema={self.schema_name}'
+        # NOTE: `maxsize=1` is important! Concurrency will be broken otherwise.
+        # NOTE: https://github.com/tortoise/tortoise-orm/issues/792
+        return f'{self.kind}://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}?schema={self.schema_name}&maxsize=1'
 
     @validator('immune_tables')
     def valid_immune_tables(cls, v):
@@ -85,6 +89,25 @@ class PostgresDatabaseConfig:
         return v
 
 
+@dataclass
+class HTTPConfig:
+    cache: Optional[bool] = None
+    retry_count: Optional[int] = None
+    retry_sleep: Optional[float] = None
+    retry_multiplier: Optional[float] = None
+    ratelimit_rate: Optional[int] = None
+    ratelimit_period: Optional[int] = None
+    connection_limit: Optional[int] = None
+    batch_size: Optional[int] = None
+
+    def merge(self, other: Optional['HTTPConfig']) -> None:
+        if not other:
+            return
+        for k, v in other.__dict__.items():
+            if v is not None:
+                setattr(self, k, v)
+
+
 @dataclass
 class ContractConfig:
     """Contract config
@@ -137,16 +160,18 @@ class TzktDatasourceConfig(NameMixin):
 
     kind: Literal['tzkt']
     url: str
+    http: Optional[HTTPConfig] = None
 
     def __hash__(self):
         return hash(self.url)
 
-    @validator('url', allow_reuse=True)
-    def valid_url(cls, v):
-        parsed_url = urlparse(v)
+    def __post_init_post_parse__(self) -> None:
+        super().__post_init_post_parse__()
+        if self.http and self.http.batch_size and self.http.batch_size > 10000:
+            raise ConfigurationError('`batch_size` must be less than 10000')
+        parsed_url = urlparse(self.url)
         if not (parsed_url.scheme and parsed_url.netloc):
-            raise ConfigurationError(f'`{v}` is not a valid datasource URL')
-        return v
+            raise ConfigurationError(f'`{self.url}` is not a valid datasource URL')
 
 
 @dataclass
@@ -159,6 +184,7 @@ class BcdDatasourceConfig(NameMixin):
     kind: Literal['bcd']
     url: str
     network: str
+    http: Optional[HTTPConfig] = None
 
     def __hash__(self):
         return hash(self.url + self.network)
@@ -171,7 +197,19 @@ class BcdDatasourceConfig(NameMixin):
         return v
 
 
-DatasourceConfigT = Union[TzktDatasourceConfig, BcdDatasourceConfig]
+@dataclass
+class CoinbaseDatasourceConfig(NameMixin):
+    kind: Literal['coinbase']
+    api_key: Optional[str] = None
+    secret_key: Optional[str] = None
+    passphrase: Optional[str] = None
+    http: Optional[HTTPConfig] = None
+
+    def __hash__(self):
+        return hash(self.kind)
+
+
+DatasourceConfigT = Union[TzktDatasourceConfig, BcdDatasourceConfig, CoinbaseDatasourceConfig]
 
 
 @dataclass
@@ -193,6 +231,7 @@ class PatternConfig(ABC):
 
     @classmethod
     def format_parameter_import(cls, package: str, module_name: str, entrypoint: str) -> str:
+        entrypoint = entrypoint.lstrip('_')
         parameter_cls = f'{snake_to_pascal(entrypoint)}Parameter'
         return f'from {package}.types.{module_name}.parameter.{pascal_to_snake(entrypoint)} import {parameter_cls}'
 
@@ -205,6 +244,7 @@ class PatternConfig(ABC):
 
     @classmethod
     def format_operation_argument(cls, module_name: str, entrypoint: str, optional: bool) -> str:
+        entrypoint = entrypoint.lstrip('_')
         parameter_cls = f'{snake_to_pascal(entrypoint)}Parameter'
         storage_cls = f'{snake_to_pascal(module_name)}Storage'
         if optional:
@@ -424,6 +464,10 @@ class HandlerConfig:
 
     def __post_init_post_parse__(self):
         self._callback_fn = None
+        if self.callback in (ROLLBACK_HANDLER, CONFIGURE_HANDLER):
+            raise ConfigurationError(f'`{self.callback}` callback name is reserved')
+        if self.callback and self.callback != pascal_to_snake(self.callback):
+            raise ConfigurationError('`callback` field must conform to snake_case naming style')
 
     @property
     def callback_fn(self) -> Callable:
@@ -473,12 +517,9 @@ class IndexConfig(TemplateValuesMixin, NameMixin):
         NameMixin.__post_init_post_parse__(self)
 
     def hash(self) -> str:
-        return hashlib.sha256(
-            json.dumps(
-                self,
-                default=pydantic_encoder,
-            ).encode(),
-        ).hexdigest()
+        config_json = json.dumps(self, default=pydantic_encoder)
+        config_hash = hashlib.sha256(config_json.encode()).hexdigest()
+        return config_hash
 
     @property
     def datasource_config(self) -> TzktDatasourceConfig:
@@ -599,13 +640,50 @@ HandlerPatternConfigT = Union[OperationHandlerOriginationPatternConfig, Operatio
 class HasuraConfig:
     url: str
     admin_secret: Optional[str] = None
+    source: str = 'default'
+    select_limit: int = 100
+    allow_aggregations: bool = True
+    camel_case: bool = False
+    connection_timeout: int = 5
+    rest: bool = True
+    http: Optional[HTTPConfig] = None
 
     @validator('url', allow_reuse=True)
     def valid_url(cls, v):
         parsed_url = urlparse(v)
         if not (parsed_url.scheme and parsed_url.netloc):
             raise ConfigurationError(f'`{v}` is not a valid Hasura URL')
-        return v
+        return v.rstrip('/')
+
+    @validator('source', allow_reuse=True)
+    def valid_source(cls, v):
+        if v != 'default':
+            raise NotImplementedError('Multiple Hasura sources are not supported at the moment')
+
+    @property
+    def headers(self) -> Dict[str, str]:
+        if self.admin_secret:
+            return {'X-Hasura-Admin-Secret': self.admin_secret}
+        return {}
+
+
+@dataclass
+class JobConfig(HandlerConfig):
+    crontab: Optional[str] = None
+    interval: Optional[int] = None
+    args: Optional[Dict[str, Any]] = None
+    atomic: bool = False
+
+    def __post_init_post_parse__(self):
+        if int(bool(self.crontab)) + int(bool(self.interval)) != 1:
+            raise ConfigurationError('Either `interval` or `crontab` field must be specified')
+        super().__post_init_post_parse__()
+
+
+@dataclass
+class SentryConfig:
+    dsn: str
+    environment: Optional[str] = None
 
 
 @dataclass
@@ -620,18 +698,32 @@ class DipDupConfig:
     :param templates: Mapping of template aliases and index templates
     :param database: Database config
     :param hasura: Hasura config
+    :param jobs: Mapping of job aliases and job configs
+    :param sentry: Sentry integration config
     """
 
     spec_version: str
     package: str
-    datasources: Dict[str, Union[TzktDatasourceConfig, BcdDatasourceConfig]]
+    datasources: Dict[str, DatasourceConfigT]
     contracts: Dict[str, ContractConfig] = Field(default_factory=dict)
     indexes: Dict[str, IndexConfigT] = Field(default_factory=dict)
     templates: Optional[Dict[str, IndexConfigTemplateT]] = None
     database: Union[SqliteDatabaseConfig, PostgresDatabaseConfig] = SqliteDatabaseConfig(kind='sqlite')
     hasura: Optional[HasuraConfig] = None
+    jobs: Optional[Dict[str, JobConfig]] = None
+    sentry: Optional[SentryConfig] = None
+
+    @property
+    def environment(self) -> Dict[str, str]:
+        return self._environment
+
+    @property
+    def filenames(self) -> List[str]:
+        return self._filenames
 
     def __post_init_post_parse__(self):
+        self._filenames: List[str] = []
+        self._environment: Dict[str, str] = {}
         self._callback_patterns: Dict[str, List[Sequence[HandlerPatternConfigT]]] = defaultdict(list)
         self._pre_initialized = []
         self._initialized = []
@@ -669,17 +761,17 @@ class DipDupConfig:
 
     def get_rollback_fn(self) -> Type:
         try:
-            module = f'{self.package}.handlers.{ROLLBACK_HANDLER}'
-            return getattr(importlib.import_module(module), ROLLBACK_HANDLER)
+            module_name = f'{self.package}.handlers.{ROLLBACK_HANDLER}'
+            return getattr(importlib.import_module(module_name), ROLLBACK_HANDLER)
         except (ModuleNotFoundError, AttributeError) as e:
-            raise HandlerImportError(f'Module `{module}` not found. Have you forgot to call `init`?') from e
+            raise HandlerImportError(module=module_name, obj=ROLLBACK_HANDLER) from e
 
     def get_configure_fn(self) -> Type:
         try:
-            module = f'{self.package}.handlers.{CONFIGURE_HANDLER}'
-            return getattr(importlib.import_module(module), CONFIGURE_HANDLER)
+            module_name = f'{self.package}.handlers.{CONFIGURE_HANDLER}'
+            return getattr(importlib.import_module(module_name), CONFIGURE_HANDLER)
         except (ModuleNotFoundError, AttributeError) as e:
-            raise HandlerImportError(f'Module `{module}` not found. Have you forgot to call `init`?') from e
+            raise HandlerImportError(module=module_name, obj=CONFIGURE_HANDLER) from e
 
     def resolve_static_templates(self) -> None:
         _logger.info('Substituting index templates')
@@ -710,10 +802,10 @@ class DipDupConfig:
                     if isinstance(contract, str):
                         index_config.contracts[i] = self.get_contract(contract)
 
-            transaction_id = 0
             for handler_config in index_config.handlers:
                 self._callback_patterns[handler_config.callback].append(handler_config.pattern)
                 for pattern_config in handler_config.pattern:
+                    transaction_id = 0
                     if isinstance(pattern_config, OperationHandlerTransactionPatternConfig):
                         if isinstance(pattern_config.destination, str):
                             pattern_config.destination = self.get_contract(pattern_config.destination)
@@ -794,6 +886,7 @@ class DipDupConfig:
         current_workdir = os.path.join(os.getcwd())
 
         json_config: Dict[str, Any] = {}
+        config_environment: Dict[str, str] = {}
         for filename in filenames:
             filename = os.path.join(current_workdir, filename)
 
@@ -804,6 +897,7 @@ class DipDupConfig:
             _logger.info('Substituting environment variables')
             for match in re.finditer(ENV_VARIABLE_REGEX, raw_config):
                 variable, default_value = match.group(1), match.group(2)
+                config_environment[variable] = default_value
                 value = env.get(variable)
                 if not default_value and not value:
                     raise ConfigurationError(f'Environment variable `{variable}` is not set')
@@ -815,14 +909,33 @@ class DipDupConfig:
                 **YAML(typ='base').load(raw_config),
             }
 
-        config = cls(**json_config)
+        try:
+            config = cls(**json_config)
+            config._environment = config_environment
+            config._filenames = filenames
+        except Exception as e:
+            raise ConfigurationError(str(e)) from e
         return config
 
     def _initialize_handler_callback(self, handler_config: HandlerConfig) -> None:
         _logger.info('Registering handler callback `%s`', handler_config.callback)
-        handler_module = importlib.import_module(f'{self.package}.handlers.{handler_config.callback}')
-        callback_fn = getattr(handler_module, handler_config.callback)
-        handler_config.callback_fn = callback_fn
+        try:
+            module_name = f'{self.package}.handlers.{handler_config.callback}'
+            module = importlib.import_module(module_name)
+            callback_fn = getattr(module, handler_config.callback)
+            handler_config.callback_fn = callback_fn
+        except (ModuleNotFoundError, AttributeError) as e:
+            raise HandlerImportError(module=module_name, obj=handler_config.callback) from e
+
+    def _initialize_job_callback(self, job_config: JobConfig) -> None:
+        _logger.info('Registering job callback `%s`', job_config.callback)
+        try:
+            module_name = f'{self.package}.jobs.{job_config.callback}'
+            module = importlib.import_module(module_name)
+            callback_fn = getattr(module, job_config.callback)
+            job_config.callback_fn = callback_fn
+        except (ModuleNotFoundError, AttributeError) as e:
+            raise HandlerImportError(module=module_name, obj=job_config.callback) from e
 
     def _initialize_index(self, index_name: str, index_config: IndexConfigT) -> None:
         if index_name in self._initialized:
@@ -856,24 +969,15 @@ class DipDupConfig:
                 self._initialize_handler_callback(big_map_handler_config)
 
                 _logger.info('Registering big map types for path `%s`', big_map_handler_config.path)
-                key_type_module = importlib.import_module(
-                    f'{self.package}'
-                    f'.types'
-                    f'.{big_map_handler_config.contract_config.module_name}'
-                    f'.big_map'
-                    f'.{pascal_to_snake(big_map_handler_config.path)}_key'
-                )
-                key_type_cls = getattr(key_type_module, snake_to_pascal(big_map_handler_config.path + '_key'))
+                module_name = big_map_handler_config.contract_config.module_name
+                big_map_path = pascal_to_snake(big_map_handler_config.path.replace('.', '_'))
+
+                key_type_module = importlib.import_module(f'{self.package}.types.{module_name}.big_map.{big_map_path}_key')
+                key_type_cls = getattr(key_type_module, snake_to_pascal(big_map_path + '_key'))
                 big_map_handler_config.key_type_cls = key_type_cls
 
-                value_type_module = importlib.import_module(
-                    f'{self.package}'
-                    f'.types'
-                    f'.{big_map_handler_config.contract_config.module_name}'
-                    f'.big_map'
-                    f'.{pascal_to_snake(big_map_handler_config.path)}_value'
-                )
-                value_type_cls = getattr(value_type_module, snake_to_pascal(big_map_handler_config.path + '_value'))
+                value_type_module = importlib.import_module(f'{self.package}.types.{module_name}.big_map.{big_map_path}_value')
+                value_type_cls = getattr(value_type_module, snake_to_pascal(big_map_path + '_value'))
                 big_map_handler_config.value_type_cls = value_type_cls
 
         else:
@@ -881,12 +985,19 @@ class DipDupConfig:
 
         self._initialized.append(index_name)
 
+    def _initialize_jobs(self) -> None:
+        if not self.jobs:
+            return
+        for job_config in self.jobs.values():
+            self._initialize_job_callback(job_config)
+
     def initialize(self) -> None:
         _logger.info('Setting up handlers and types for package `%s`', self.package)
 
         self.pre_initialize()
         for index_name, index_config in self.indexes.items():
             self._initialize_index(index_name, index_config)
+        self._initialize_jobs()
 
 
 @dataclass
