@@ -1,5 +1,6 @@
 import os
 import sys
+from pprint import pformat
 from typing import Any, Dict, Optional
 
 from tortoise import Tortoise
@@ -7,7 +8,7 @@ from tortoise.transactions import in_transaction
 
 from dipdup.config import ContractConfig, DipDupConfig, PostgresDatabaseConfig, StaticTemplateConfig
 from dipdup.datasources import DatasourceT
-from dipdup.exceptions import ConfigurationError
+from dipdup.exceptions import ContractAlreadyExistsError, IndexAlreadyExistsError
 from dipdup.utils import FormattedLogger
 
 
@@ -22,6 +23,9 @@ class DipDupContext:
         self.config = config
         self._updated: bool = False
 
+    def __str__(self) -> str:
+        return pformat(self.__dict__)
+
     def commit(self) -> None:
         """Spawn indexes after handler execution"""
         self._updated = True
@@ -83,7 +87,7 @@ class HandlerContext(DipDupContext):
 
     def add_contract(self, name: str, address: str, typename: Optional[str] = None) -> None:
         if name in self.config.contracts:
-            raise ConfigurationError(f'Contract `{name}` is already exists')
+            raise ContractAlreadyExistsError(self, name, address)
         self.config.contracts[name] = ContractConfig(
             address=address,
             typename=typename,
@@ -92,7 +96,7 @@ class HandlerContext(DipDupContext):
 
     def add_index(self, name: str, template: str, values: Dict[str, Any]) -> None:
         if name in self.config.indexes:
-            raise ConfigurationError(f'Index `{name}` is already exists')
+            raise IndexAlreadyExistsError(self, name)
         self.config.get_template(template)
         self.config.indexes[name] = StaticTemplateConfig(
             template=template,
