@@ -1,6 +1,7 @@
 from typing import Union
 
 from dipdup.datasources.bcd.datasource import BcdDatasource
+from dipdup.datasources.coinbase.datasource import CoinbaseDatasource
 from dipdup.datasources.tzkt.datasource import TzktDatasource
 
-DatasourceT = Union[TzktDatasource, BcdDatasource]
+DatasourceT = Union[TzktDatasource, BcdDatasource, CoinbaseDatasource]
