@@ -1,5 +1,7 @@
 import asyncio
 import logging
+from datetime import datetime, timezone
+from decimal import Decimal
 from enum import Enum
 from typing import Any, AsyncGenerator, Dict, List, NoReturn, Optional, Set, Tuple, cast
 
@@ -11,19 +13,18 @@ from aiosignalrcore.transport.websockets.connection import ConnectionState  # ty
 from dipdup.config import (
     BigMapIndexConfig,
     ContractConfig,
+    HTTPConfig,
     IndexConfigTemplateT,
     OperationHandlerOriginationPatternConfig,
     OperationIndexConfig,
 )
 from dipdup.datasources.datasource import IndexDatasource
-from dipdup.datasources.proxy import DatasourceRequestProxy
 from dipdup.datasources.tzkt.enums import TzktMessageType
-from dipdup.models import BigMapAction, BigMapData, OperationData
+from dipdup.models import BigMapAction, BigMapData, BlockData, HeadBlockData, OperationData
 from dipdup.utils import split_by_chunks
 
 OperationID = int
 
-TZKT_HTTP_REQUEST_LIMIT = 10000
 TZKT_ORIGINATIONS_REQUEST_LIMIT = 100
 OPERATION_FIELDS = (
     "type",
@@ -35,6 +36,7 @@ OPERATION_FIELDS = (
     "sender",
     "nonce",
     "target",
+    "initiator",
     "amount",
     "storage",
     "status",
@@ -80,14 +82,16 @@ class OperationFetcher:
         last_level: int,
         transaction_addresses: Set[str],
         origination_addresses: Set[str],
+        cache: bool = False,
     ) -> None:
         self._datasource = datasource
         self._first_level = first_level
         self._last_level = last_level
         self._transaction_addresses = transaction_addresses
         self._origination_addresses = origination_addresses
+        self._cache = cache
 
-        self._logger = logging.getLogger('dipdup.tzkt.fetcher')
+        self._logger = logging.getLogger('dipdup.tzkt')
         self._head: int = 0
         self._heads: Dict[OperationFetcherChannel, int] = {}
         self._offsets: Dict[OperationFetcherChannel, int] = {}
@@ -117,6 +121,7 @@ class OperationFetcher:
             offset=self._offsets[key],
             first_level=self._first_level,
             last_level=self._last_level,
+            cache=self._cache,
         )
 
         for op in originations:
@@ -151,6 +156,7 @@ class OperationFetcher:
             offset=self._offsets[key],
             first_level=self._first_level,
             last_level=self._last_level,
+            cache=self._cache,
         )
 
         for op in transactions:
@@ -211,14 +217,15 @@ class BigMapFetcher:
         last_level: int,
         big_map_addresses: Set[str],
         big_map_paths: Set[str],
+        cache: bool = False,
     ) -> None:
+        self._logger = logging.getLogger('dipdup.tzkt')
         self._datasource = datasource
         self._first_level = first_level
         self._last_level = last_level
         self._big_map_addresses = big_map_addresses
         self._big_map_paths = big_map_paths
-
-        self._logger = logging.getLogger('dipdup.tzkt.fetcher')
+        self._cache = cache
 
     async def fetch_big_maps_by_level(self) -> AsyncGenerator[Tuple[int, List[BigMapData]], None]:
         """Fetch big map diffs via Fetcher (not implemented yet) and pass to message callback"""
@@ -233,6 +240,7 @@ class BigMapFetcher:
                 offset,
                 self._first_level,
                 self._last_level,
+                cache=self._cache,
             )
             big_maps += fetched_big_maps
 
@@ -263,24 +271,26 @@ class TzktDatasource(IndexDatasource):
     * Calls Matchers to match received operation groups with indexes' pattern and spawn callbacks on match
     """
 
-    def __init__(self, url: str, cache: bool) -> None:
-        super().__init__()
-        self._url = url.rstrip('/')
-
+    def __init__(
+        self,
+        url: str,
+        http_config: Optional[HTTPConfig] = None,
+    ) -> None:
+        super().__init__(url, http_config)
         self._logger = logging.getLogger('dipdup.tzkt')
         self._transaction_subscriptions: Set[str] = set()
         self._origination_subscriptions: bool = False
         self._big_map_subscriptions: Dict[str, List[str]] = {}
 
         self._client: Optional[BaseHubConnection] = None
-        self._proxy = DatasourceRequestProxy(cache)
 
+        self._block: Optional[HeadBlockData] = None
         self._level: Optional[int] = None
         self._sync_level: Optional[int] = None
 
     @property
     def request_limit(self) -> int:
-        return TZKT_HTTP_REQUEST_LIMIT
+        return self._http_config.batch_size or 10000
 
     @property
     def level(self) -> Optional[int]:
@@ -290,17 +300,20 @@ class TzktDatasource(IndexDatasource):
     def sync_level(self) -> Optional[int]:
         return self._sync_level
 
-    async def close_session(self) -> None:
-        await self._proxy.close_session()
+    @property
+    def block(self) -> HeadBlockData:
+        if self._block is None:
+            raise RuntimeError('No message from `head` channel received')
+        return self._block
 
     async def get_similar_contracts(self, address: str, strict: bool = False) -> List[str]:
         """Get list of contracts sharing the same code hash or type hash"""
         entrypoint = 'same' if strict else 'similar'
         self._logger.info('Fetching %s contracts for address `%s', entrypoint, address)
 
-        contracts = await self._proxy.http_request(
+        contracts = await self._http.request(
             'get',
-            url=f'{self._url}/v1/contracts/{address}/{entrypoint}',
+            url=f'v1/contracts/{address}/{entrypoint}',
             params=dict(
                 select='address',
                 limit=self.request_limit,
@@ -311,62 +324,71 @@ class TzktDatasource(IndexDatasource):
     async def get_originated_contracts(self, address: str) -> List[str]:
         """Get contracts originated from given address"""
         self._logger.info('Fetching originated contracts for address `%s', address)
-        contracts = await self._proxy.http_request(
+        contracts = await self._http.request(
             'get',
-            url=f'{self._url}/v1/accounts/{address}/contracts',
+            url=f'v1/accounts/{address}/contracts',
             params=dict(
                 limit=self.request_limit,
             ),
-            skip_cache=True,
         )
         return [c['address'] for c in contracts]
 
     async def get_contract_summary(self, address: str) -> Dict[str, Any]:
         """Get contract summary"""
         self._logger.info('Fetching contract summary for address `%s', address)
-        return await self._proxy.http_request(
+        return await self._http.request(
             'get',
-            url=f'{self._url}/v1/contracts/{address}',
+            url=f'v1/contracts/{address}',
         )
 
     async def get_contract_storage(self, address: str) -> Dict[str, Any]:
         """Get contract storage"""
         self._logger.info('Fetching contract storage for address `%s', address)
-        return await self._proxy.http_request(
+        return await self._http.request(
             'get',
-            url=f'{self._url}/v1/contracts/{address}/storage',
+            url=f'v1/contracts/{address}/storage',
         )
 
     async def get_jsonschemas(self, address: str) -> Dict[str, Any]:
         """Get JSONSchemas for contract's storage/parameter/bigmap types"""
         self._logger.info('Fetching jsonschemas for address `%s', address)
-        jsonschemas = await self._proxy.http_request(
+        jsonschemas = await self._http.request(
             'get',
-            url=f'{self._url}/v1/contracts/{address}/interface',
+            url=f'v1/contracts/{address}/interface',
+            cache=True,
         )
         self._logger.debug(jsonschemas)
         return jsonschemas
 
-    async def get_latest_block(self) -> Dict[str, Any]:
+    async def get_head_block(self) -> HeadBlockData:
         """Get latest block (head)"""
         self._logger.info('Fetching latest block')
-        block = await self._proxy.http_request(
+        head_block_json = await self._http.request(
             'get',
-            url=f'{self._url}/v1/head',
-            skip_cache=True,
+            url='v1/head',
         )
-        self._logger.debug(block)
-        return block
+        return self.convert_head_block(head_block_json)
 
-    async def get_originations(self, addresses: Set[str], offset: int, first_level: int, last_level: int) -> List[OperationData]:
+    async def get_block(self, level: int) -> BlockData:
+        """Get block by level"""
+        self._logger.info('Fetching block %s', level)
+        block_json = await self._http.request(
+            'get',
+            url=f'v1/blocks/{level}',
+        )
+        return self.convert_block(block_json)
+
+    async def get_originations(
+        self, addresses: Set[str], offset: int, first_level: int, last_level: int, cache: bool = False
+    ) -> List[OperationData]:
         raw_originations = []
         # NOTE: TzKT may hit URL length limit with hundreds of originations in a single request.
         # NOTE: Chunk of 100 addresses seems like a reasonable choice - URL of ~3971 characters.
         # NOTE: Other operation requests won't hit that limit.
         for addresses_chunk in split_by_chunks(list(addresses), TZKT_ORIGINATIONS_REQUEST_LIMIT):
-            raw_originations += await self._proxy.http_request(
+            raw_originations += await self._http.request(
                 'get',
-                url=f'{self._url}/v1/operations/originations',
+                url='v1/operations/originations',
                 params={
                     "originatedContract.in": ','.join(addresses_chunk),
                     "offset": offset,
@@ -376,6 +398,7 @@ class TzktDatasource(IndexDatasource):
                     "select": ','.join(ORIGINATION_OPERATION_FIELDS),
                     "status": "applied",
                 },
+                cache=cache,
             )
 
         originations = []
@@ -386,11 +409,11 @@ class TzktDatasource(IndexDatasource):
         return originations
 
     async def get_transactions(
-        self, field: str, addresses: Set[str], offset: int, first_level: int, last_level: int
+        self, field: str, addresses: Set[str], offset: int, first_level: int, last_level: int, cache: bool = False
     ) -> List[OperationData]:
-        raw_transactions = await self._proxy.http_request(
+        raw_transactions = await self._http.request(
             'get',
-            url=f'{self._url}/v1/operations/transactions',
+            url='v1/operations/transactions',
             params={
                 f"{field}.in": ','.join(addresses),
                 "offset": offset,
@@ -400,6 +423,7 @@ class TzktDatasource(IndexDatasource):
                 "select": ','.join(TRANSACTION_OPERATION_FIELDS),
                 "status": "applied",
             },
+            cache=cache,
         )
         transactions = []
         for op in raw_transactions:
@@ -408,10 +432,12 @@ class TzktDatasource(IndexDatasource):
             transactions.append(self.convert_operation(op))
         return transactions
 
-    async def get_big_maps(self, addresses: Set[str], paths: Set[str], offset: int, first_level: int, last_level: int) -> List[BigMapData]:
-        raw_big_maps = await self._proxy.http_request(
+    async def get_big_maps(
+        self, addresses: Set[str], paths: Set[str], offset: int, first_level: int, last_level: int, cache: bool = False
+    ) -> List[BigMapData]:
+        raw_big_maps = await self._http.request(
             'get',
-            url=f'{self._url}/v1/bigmaps/updates',
+            url='v1/bigmaps/updates',
             params={
                 "contract.in": ",".join(addresses),
                 "paths.in": ",".join(paths),
@@ -420,6 +446,7 @@ class TzktDatasource(IndexDatasource):
                 "level.gt": first_level,
                 "level.le": last_level,
             },
+            cache=cache,
         )
         big_maps = []
         for bm in raw_big_maps:
@@ -460,7 +487,7 @@ class TzktDatasource(IndexDatasource):
             self._logger.info('Creating websocket client')
             self._client = (
                 HubConnectionBuilder()
-                .with_url(self._url + '/v1/events')
+                .with_url(self._http._url + '/v1/events')
                 .with_automatic_reconnect(
                     {
                         "type": "raw",
@@ -475,6 +502,7 @@ class TzktDatasource(IndexDatasource):
             self._client.on_error(self._on_error)
             self._client.on('operations', self._on_operation_message)
             self._client.on('bigmaps', self._on_big_map_message)
+            self._client.on('head', self._on_head_message)
 
         return self._client
 
@@ -491,6 +519,7 @@ class TzktDatasource(IndexDatasource):
             return
 
         self._logger.info('Connected to server')
+        await self.subscribe_to_head()
         for address in self._transaction_subscriptions:
             await self.subscribe_to_transactions(address)
         # NOTE: All originations are passed to matcher
@@ -506,11 +535,7 @@ class TzktDatasource(IndexDatasource):
     async def subscribe_to_transactions(self, address: str) -> None:
         """Subscribe to contract's operations on established WS connection"""
         self._logger.info('Subscribing to %s transactions', address)
-
-        while self._get_client().transport.state != ConnectionState.connected:
-            await asyncio.sleep(0.1)
-
-        await self._get_client().send(
+        await self._send(
             'SubscribeToOperations',
             [
                 {
@@ -523,11 +548,7 @@ class TzktDatasource(IndexDatasource):
     async def subscribe_to_originations(self) -> None:
         """Subscribe to all originations on established WS connection"""
         self._logger.info('Subscribing to originations')
-
-        while self._get_client().transport.state != ConnectionState.connected:
-            await asyncio.sleep(0.1)
-
-        await self._get_client().send(
+        await self._send(
             'SubscribeToOperations',
             [
                 {
@@ -539,12 +560,8 @@ class TzktDatasource(IndexDatasource):
     async def subscribe_to_big_maps(self, address: str, paths: List[str]) -> None:
         """Subscribe to contract's big map diffs on established WS connection"""
         self._logger.info('Subscribing to big map updates of %s, %s', address, paths)
-
-        while self._get_client().transport.state != ConnectionState.connected:
-            await asyncio.sleep(0.1)
-
         for path in paths:
-            await self._get_client().send(
+            await self._send(
                 'SubscribeToBigMaps',
                 [
                     {
@@ -554,12 +571,26 @@ class TzktDatasource(IndexDatasource):
                 ],
             )
 
-    async def _on_operation_message(
-        self,
-        message: List[Dict[str, Any]],
-    ) -> None:
-        """Parse and emit raw operations from WS"""
+    async def subscribe_to_head(self) -> None:
+        """Subscribe to head on established WS connection"""
+        self._logger.info('Subscribing to head')
+        await self._send(
+            'SubscribeToHead',
+            [],
+        )
 
+    def _default_http_config(self) -> HTTPConfig:
+        return HTTPConfig(
+            cache=True,
+            retry_sleep=1,
+            retry_multiplier=1.1,
+            ratelimit_rate=100,
+            ratelimit_period=30,
+            connection_limit=25,
+        )
+
+    async def _on_operation_message(self, message: List[Dict[str, Any]]) -> None:
+        """Parse and emit raw operations from WS"""
         for item in message:
             current_level = item['state']
             message_type = TzktMessageType(item['type'])
@@ -577,7 +608,8 @@ class TzktDatasource(IndexDatasource):
                     if operation.status != 'applied':
                         continue
                     operations.append(operation)
-                self.emit_operations(operations)
+                if operations:
+                    self.emit_operations(operations, self.block)
 
             elif message_type == TzktMessageType.REORG:
                 if self.level is None:
@@ -587,10 +619,7 @@ class TzktDatasource(IndexDatasource):
             else:
                 raise NotImplementedError
 
-    async def _on_big_map_message(
-        self,
-        message: List[Dict[str, Any]],
-    ) -> None:
+    async def _on_big_map_message(self, message: List[Dict[str, Any]]) -> None:
         """Parse and emit raw big map diffs from WS"""
         for item in message:
             current_level = item['state']
@@ -617,6 +646,31 @@ class TzktDatasource(IndexDatasource):
             else:
                 raise NotImplementedError
 
+    async def _on_head_message(self, message: List[Dict[str, Any]]) -> None:
+        for item in message:
+            current_level = item['state']
+            message_type = TzktMessageType(item['type'])
+            self._logger.info('Got block message, %s, level %s', message_type, current_level)
+
+            if message_type == TzktMessageType.STATE:
+                self._sync_level = current_level
+                self._level = current_level
+
+            elif message_type == TzktMessageType.DATA:
+                self._level = current_level
+                block_json = item['data']
+                block = self.convert_head_block(block_json)
+                self._block = block
+                self.emit_head(block)
+
+            elif message_type == TzktMessageType.REORG:
+                if self.level is None:
+                    raise RuntimeError
+                self.emit_rollback(self.level, current_level)
+
+            else:
+                raise NotImplementedError
+
     @classmethod
     def convert_operation(cls, operation_json: Dict[str, Any]) -> OperationData:
         """Convert raw operation message from WS/REST into dataclass"""
@@ -629,18 +683,20 @@ class TzktDatasource(IndexDatasource):
             type=operation_json['type'],
             id=operation_json['id'],
             level=operation_json['level'],
-            timestamp=operation_json['timestamp'],
+            timestamp=cls._parse_timestamp(operation_json['timestamp']),
             block=operation_json.get('block'),
             hash=operation_json['hash'],
             counter=operation_json['counter'],
             sender_address=operation_json['sender']['address'] if operation_json.get('sender') else None,
             target_address=operation_json['target']['address'] if operation_json.get('target') else None,
-            amount=operation_json.get('amount'),
+            initiator_address=operation_json['initiator']['address'] if operation_json.get('initiator') else None,
+            amount=operation_json.get('amount') or operation_json.get('contractBalance'),
             status=operation_json['status'],
             has_internals=operation_json.get('hasInternals'),
             sender_alias=operation_json['sender'].get('alias'),
             nonce=operation_json.get('nonce'),
             target_alias=operation_json['target'].get('alias') if operation_json.get('target') else None,
+            initiator_alias=operation_json['initiator'].get('alias') if operation_json.get('initiator') else None,
             entrypoint=operation_json['parameter']['entrypoint'] if operation_json.get('parameter') else None,
             parameter_json=operation_json['parameter']['value'] if operation_json.get('parameter') else None,
             originated_contract_address=operation_json['originatedContract']['address']
@@ -664,7 +720,7 @@ class TzktDatasource(IndexDatasource):
             level=big_map_json['level'],
             # FIXME: operation_id field in API
             operation_id=big_map_json['level'],
-            timestamp=big_map_json['timestamp'],
+            timestamp=cls._parse_timestamp(big_map_json['timestamp']),
             bigmap=big_map_json['bigmap'],
             contract_address=big_map_json['contract']['address'],
             path=big_map_json['path'],
@@ -672,3 +728,55 @@ class TzktDatasource(IndexDatasource):
             key=big_map_json.get('content', {}).get('key'),
             value=big_map_json.get('content', {}).get('value'),
         )
+
+    @classmethod
+    def convert_block(cls, block_json: Dict[str, Any]) -> BlockData:
+        """Convert raw block message from REST into dataclass"""
+        return BlockData(
+            level=block_json['level'],
+            hash=block_json['hash'],
+            timestamp=cls._parse_timestamp(block_json['timestamp']),
+            proto=block_json['proto'],
+            priority=block_json['priority'],
+            validations=block_json['validations'],
+            deposit=block_json['deposit'],
+            reward=block_json['reward'],
+            fees=block_json['fees'],
+            nonce_revealed=block_json['nonceRevealed'],
+            baker_address=block_json.get('baker', {}).get('address'),
+            baker_alias=block_json.get('baker', {}).get('alias'),
+        )
+
+    @classmethod
+    def convert_head_block(cls, head_block_json: Dict[str, Any]) -> HeadBlockData:
+        """Convert raw head block message from WS/REST into dataclass"""
+        return HeadBlockData(
+            cycle=head_block_json['cycle'],
+            level=head_block_json['level'],
+            hash=head_block_json['hash'],
+            protocol=head_block_json['protocol'],
+            timestamp=cls._parse_timestamp(head_block_json['timestamp']),
+            voting_epoch=head_block_json['votingEpoch'],
+            voting_period=head_block_json['votingPeriod'],
+            known_level=head_block_json['knownLevel'],
+            last_sync=head_block_json['lastSync'],
+            synced=head_block_json['synced'],
+            quote_level=head_block_json['quoteLevel'],
+            quote_btc=Decimal(head_block_json['quoteBtc']),
+            quote_eur=Decimal(head_block_json['quoteEur']),
+            quote_usd=Decimal(head_block_json['quoteUsd']),
+            quote_cny=Decimal(head_block_json['quoteCny']),
+            quote_jpy=Decimal(head_block_json['quoteJpy']),
+            quote_krw=Decimal(head_block_json['quoteKrw']),
+            quote_eth=Decimal(head_block_json['quoteEth']),
+        )
+
+    async def _send(self, method: str, arguments: List[Dict[str, Any]], on_invocation=None) -> None:
+        client = self._get_client()
+        while client.transport.state != ConnectionState.connected:
+            await asyncio.sleep(0.1)
+        await client.send(method, arguments, on_invocation)
+
+    @classmethod
+    def _parse_timestamp(cls, timestamp: str) -> datetime:
+        return datetime.fromisoformat(timestamp[:-1]).replace(tzinfo=timezone.utc)
